from threading import local, Lock
from typing import Any, Callable, Optional, Type, TypeVar, Union

T = TypeVar("T")


# noinspection PyPep8Naming
class thread_local_cached_property:
    def __init__(self, func: Callable[[Any], T]):
        self.func: Callable[[Any], T] = func
        self.attr_name: Optional[str] = None
        self.__doc__: Optional[str] = func.__doc__
        self.lock: Lock = Lock()

    def __set_name__(self, owner: Type[Any], name: str) -> None:
        if self.attr_name is None:
            self.attr_name = name
        elif name != self.attr_name:
            raise TypeError(
                f"Cannot assign the same {type(self).__name__} to two different names "
                f"({self.attr_name!r} and {name!r})."
            )

    def __get__(
        self, instance: Any, owner: Optional[Type[Any]] = None
    ) -> Union[T, "thread_local_cached_property"]:
        if instance is None:
            return self
        if self.attr_name is None:
            raise TypeError(
                f"Cannot use {type(self).__name__}  instance without calling __set_name__ on it."
            )

        with self.lock:
            if not hasattr(instance, "__local_dict__"):
                instance.__local_dict__ = local()
            cache = instance.__local_dict__

            if not hasattr(cache, self.attr_name):
                setattr(cache, self.attr_name, self.func(instance))

        return getattr(cache, self.attr_name)



import pytest

from autowired import thread_local_cached_property


def test_thread_local_cached_property_value():
    next_id = 0

    class TestClass:
        @thread_local_cached_property
        def value(self):
            nonlocal next_id
            next_id += 1
            return next_id

    instance = TestClass()
    assert instance.value == 1
    assert instance.value == 1

    def in_new_thread():
        assert instance.value == 2
        assert instance.value == 2

    import threading

    thread = threading.Thread(target=in_new_thread)
    thread.start()
    thread.join()

    assert instance.value == 1


def test_double_assignment_thread_local_cached_property():
    class MyClass:
        pass

    def f():
        pass

    descriptor = thread_local_cached_property(f)

    descriptor.__set_name__(MyClass, "attr1")

    with pytest.raises(TypeError) as exc_info:
        descriptor.__set_name__(MyClass, "attr2")


def test_thread_local_cached_property_without_set_name():
    class TestClass:
        pass

    test_instance = TestClass()

    prop_without_set_name = thread_local_cached_property(lambda x: x)

    with pytest.raises(TypeError):
        prop_without_set_name.__get__(test_instance)


def test_access_class_property():
    class TestClass:
        @thread_local_cached_property
        def value(self):
            return 1

    assert isinstance(TestClass.value, thread_local_cached_property)
    assert TestClass().value == 1



from abc import ABC


class AutowiredException(Exception, ABC):
    """
    Base class for all library exceptions.
    """

    pass


class MissingTypeAnnotation(AutowiredException):
    """
    Raised when a field or property is not annotated with a type hint.
    """

    pass


class IllegalAutoWireType(AutowiredException):
    """
    Raised when an object of a type that is not allowed to be auto-wired is auto-wired.
    """

    pass


class AmbiguousDependencyException(AutowiredException):
    """
    Raised when a dependency cannot be resolved because multiple candidates are found
    and none of them matches the name of the dependency.
    """

    pass


class UnresolvableDependencyException(AutowiredException):
    """
    Raised when a dependency cannot be resolved.
    """

    pass


class InstantiationError(AutowiredException):
    """
    Raised when an object cannot be instantiated.
    """

    pass


class ContextError(AutowiredException, ABC):
    """
    Base class for all context-related errors.
    """

    pass


class IllegalContextClass(ContextError):
    """
    Raised when a class is used as a context class not allowed to be used as such.
    """

    pass


class NotProvidedException(ContextError):
    """
    Raised when a provided field is not initialized.
    """

    pass



from autowired import component
from tests.component_scan_test_module.controller import Controller


@component
class FileComponent(Controller):
    pass


@component
class FileComponentInitExposed(Controller):
    pass


class FileNotComponent(Controller):
    pass



from autowired import component
from tests.component_scan_test_module import Controller


@component
class TestPackageFileComponent(Controller):
    pass


@component
class TestPackageFileComponentInitExposed(Controller):
    pass


@component(transient=True)
class TransientComponent(Controller):
    pass


class TestPackageFileNotComponent(Controller):
    pass



from typing import List

import pytest

import tests.component_scan_test_module
from autowired import Container, Dependency
from tests.component_scan_test_module import Controller


def test_component_scan():
    container = Container()

    container.component_scan(tests.component_scan_test_module)

    providers = container.get_providers()
    assert len(providers) == 7

    components = container.resolve(Dependency("components", List[Controller]))

    assert len(components) == 7

    component_class_names = set(type(c).__name__ for c in components)

    assert component_class_names == {
        "FileComponent",
        "FileComponentInitExposed",
        "RootModuleComponent",
        "TestPackageRootComponent",
        "TestPackageFileComponent",
        "TestPackageFileComponentInitExposed",
        "TransientComponent",
    }

    for provider in providers:
        expect_singleton = provider.get_name() != "transient_component"

        instance1 = provider.get_instance(Dependency("instance1", object), container)
        instance2 = provider.get_instance(Dependency("instance2", object), container)

        assert (instance1 is instance2) == expect_singleton


def test_component_scan_invalid_module():
    container = Container()

    with pytest.raises(TypeError):
        container.component_scan("tests.component_scan_test_module")



from autowired import component
from tests.component_scan_test_module.controller import Controller
from tests.component_scan_test_module.file_module import FileComponentInitExposed


@component
class RootModuleComponent(Controller):
    pass


class RootModuleNotComponent(Controller):
    pass



import importlib
import inspect
import pkgutil
import sys
from dataclasses import dataclass
from typing import Type, Iterable, Set, Optional

Module = type(sys)


@dataclass
class ClassComponentInfo:
    cls: Type
    transient: bool = False


def component(cls=None, *, transient: bool = False):
    def wrap_class(cls_to_wrap):
        cls_to_wrap._component_info = ClassComponentInfo(cls_to_wrap, transient)
        return cls_to_wrap

    if cls is None:
        # The decorator has been called like @component(transient=True)
        # Return the wrapper function to apply later with the class
        return wrap_class
    else:
        # The decorator has been called like @component without parentheses
        # Apply the wrapper function directly to the class
        return wrap_class(cls)


def get_component_info(cls) -> Optional[ClassComponentInfo]:
    return getattr(cls, "_component_info", None)


class ClassScanner:
    def __init__(self, root_module: Module):
        if not inspect.ismodule(root_module):
            raise TypeError(f"Expected a module, got {type(root_module)}")
        self.root_module = root_module

    def _get_classes(self) -> Iterable[Type]:
        for name, cls in inspect.getmembers(self.root_module, inspect.isclass):
            if cls.__module__ == self.root_module.__name__:
                yield cls

        path = self.root_module.__path__
        prefix = self.root_module.__name__ + "."

        for importer, modname, is_pkg in pkgutil.walk_packages(path, prefix):
            sub_module = importlib.import_module(modname)
            for name, cls in inspect.getmembers(sub_module, inspect.isclass):
                if cls.__module__ == sub_module.__name__:
                    yield cls

    def get_classes(self) -> Iterable[Type]:
        seen: Set[str] = set()

        def cls_to_key(cls: Type) -> str:
            return f"{cls.__module__}.{cls.__name__}"

        for cls in self._get_classes():
            key = cls_to_key(cls)
            if key not in seen:
                seen.add(key)
                yield cls


def component_scan(root_module: Module) -> Iterable[ClassComponentInfo]:
    scanner = ClassScanner(root_module)
    component_infos = (get_component_info(cls) for cls in scanner.get_classes())
    return (c for c in component_infos if c is not None)



import dataclasses
import inspect
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from types import FunctionType
from typing import Type, Callable, Any, List, Optional, Union, Generic, Dict, TypeVar, cast
from ._component_scan import component_scan, Module

from ._exceptions import (
    MissingTypeAnnotation,
    AmbiguousDependencyException,
    IllegalAutoWireType,
    InstantiationError,
    UnresolvableDependencyException,
    AutowiredException,
)
from ._logging import logger
from ._typing_utils import is_subtype, get_sequence_type

_T = TypeVar("_T")


@dataclass(frozen=True)
class Dependency:
    """
    A dependency specification.
    """

    name: str
    type: Type[_T]
    required: bool = True
    default_factory: Optional[Callable[[], _T]] = None


class Provider(ABC, Generic[_T]):
    @abstractmethod
    def get_instance(
            self, dependency: Dependency, container: "Container"
    ) -> _T:  # pragma: no cover
        """
        Returns an instance that satisfies the given dependency specification.

        :param dependency: The dependency specification.
        :param container: The container that is currently resolving the dependency.
        :return: An instance that satisfies the given dependency specification
        """
        ...

    @abstractmethod
    def get_name(self) -> str:  # pragma: no cover
        """
        Returns the name of the provider.
        Used by the container to resolve ambiguous dependencies.
        If a container contains multiple dependencies that satisfy the same dependency specification,
        the name of the dependency is compared to the provider name to try to resolve the ambiguity.

        :return: The name of the provider
        """
        ...

    @abstractmethod
    def satisfies(self, dependency: Dependency) -> bool:  # pragma: no cover
        """
        Returns whether this provider satisfies the given dependency specification.

        :param dependency: The dependency specification.
        :return: Whether this provider satisfies the given dependency specification
        """
        ...

    @staticmethod
    def from_instance(instance: _T, name: str = None) -> "Provider[_T]":
        """
        Creates a singleton provider from the given instance.

        :param instance: The instance. Will always be returned by self.get_instance(...)
        :param name: The name of the provider. If None, the type name of the instance is used (snake case).
        :return: The newly created provider
        """
        if name is None:
            name = _camel_to_snake(type(instance).__name__)
        return _SimpleProvider(name, type(instance), lambda: instance)

    # noinspection PyShadowingBuiltins
    @staticmethod
    def from_supplier(
            supplier: Callable[[], _T],
            type: Optional[Type[_T]] = None,
            name: Optional[str] = None,
    ) -> "Provider[_T]":
        """
        Creates a provider from the given supplier function.
        :param supplier: The supplier function. Will be called every time self.get_instance(...) is called.
        :param type: The type of the component this provider provides.
                        If None, the return type of the supplier function is used, or if supplier is a class,
                        the class itself is used.
        :param name: The name of the provider. If None, the type name of the supplier is used (snake case).
        :return: The newly created provider
        """
        if type is None:
            # if getter is a class, use the class as a type
            if inspect.isclass(supplier):
                type = supplier
            else:
                type = inspect.signature(supplier).return_annotation
                if type == inspect.Signature.empty:
                    raise MissingTypeAnnotation(
                        f"Failed to determine type of {supplier.__name__}. "
                    )

        if name is None:
            name = _camel_to_snake(type.__name__)

        return _SimpleProvider(name, type, supplier)

    @staticmethod
    def from_class(cls, container: "Container", transient: bool) -> "Provider[_T]":
        def supplier():
            return container.autowire(cls)

        if not transient:
            supplier = _cached(supplier)

        return _SimpleProvider(_camel_to_snake(cls.__name__), cls, supplier)


def _cached(supplier: Callable[[], _T]) -> Callable[[], _T]:
    cached = False
    result = None

    def wrapper():
        nonlocal cached
        nonlocal result
        if not cached:
            result = supplier()
            cached = True
        return result

    return wrapper


@dataclass(frozen=True)
class _SimpleProvider(Provider[_T]):
    name: str
    type: Type[_T]
    getter: Callable[[], _T] = dataclasses.field(repr=False)

    def get_instance(self, dependency: Dependency, container: "Container") -> _T:
        return self.getter()

    def get_name(self) -> str:
        return self.name

    def satisfies(self, dependency: Dependency) -> bool:
        return is_subtype(self.type, dependency.type)


_illegal_autowiredType_modules = ["builtins", "typing", "dataclasses", "abc", "object"]


def _is_illegal_type(t: Type[_T]) -> bool:
    return t.__module__.split(".")[0] in _illegal_autowiredType_modules


class Container:
    """
    A container for resolving and storing dependencies.
    """

    _providers: List[Provider]

    def __init__(self):
        self._providers = []

    def get_providers(self, dependency: Optional[Dependency] = None) -> List[Provider]:
        """
        Returns all providers that match the given dependency specification.

        :param dependency: Optional dependency specification, if None, all providers are returned
        :return:
        """

        if dependency is None:
            return list(self._providers)
        else:
            return [p for p in self._providers if p.satisfies(dependency)]

    def get_provider(self, dependency: Dependency) -> Optional[Provider]:
        """
        Returns an existing provider that matches the given dependency specification.

        :param dependency:
        :return:
        :raises AmbiguousDependencyException: If multiple matching providers are found and there is no name match
        """
        candidates = self.get_providers(dependency)

        if len(candidates) == 1:
            return candidates[0]
        if len(candidates) > 1:
            by_name = _group_by(lambda obj: obj.name, candidates)
            if dependency.name in by_name and len(by_name[dependency.name]) == 1:
                return by_name[dependency.name][0]
            else:
                raise AmbiguousDependencyException(
                    f"Failed to resolve dependency {dependency.name}"
                    f" of type {dependency.type.__name__}."
                    f" Multiple candidates found: {candidates}"
                )

        return None

    def add(self, provider_or_instance: Union[Provider, Any], /):
        """
        Adds a provider or instance (as singleton provider) to the container.

        :param provider_or_instance: If not a provider, a singleton provider is created from the instance.
                                     The name of the provider is derived from the type name of the instance.
        """
        if not isinstance(provider_or_instance, Provider):
            provider = Provider.from_instance(provider_or_instance)
        else:
            provider = provider_or_instance

        self._providers.append(provider)

    def remove(self, provider: Union[str, Provider, Type[_T]], /):
        """
        Remove a provider from the container.

        :param provider: Provider name or provider instance
        """

        def predicate(p: Provider) -> bool:
            if isinstance(provider, Provider):
                return p == provider
            else:
                return p.get_name() == provider

        remove_index = None
        for i, p in enumerate(self._providers):
            if predicate(p):
                remove_index = i
                break

        if remove_index is not None:
            self._providers.pop(remove_index)

    def resolve(self, dependency: Union[Dependency, Type[_T]]) -> _T:
        """
        Resolves a dependency from the container.
        If no existing provider satisfies the dependency specification,
        the container tries to auto-wire the object as defined by `self.autowire(...)`
        and stores the result instance as a new singleton provider.
        The same is true for the dependencies of the object (recursively).
        If multiple matching providers are found,
        the name of the dependency is compared to the provider name to try to resolve the ambiguity.

        :param dependency: Dependency specification or target type
        :return: the resolved dependency
        :raises UnresolvableDependencyException: if the dependency cannot be resolved
        :raises AmbiguousDependencyException: if multiple matching providers are found and there is no name match
        """
        if not isinstance(dependency, Dependency):
            logger.trace(f"Resolving type {dependency.__name__} for container {self}")
            dependency = Dependency(
                _camel_to_snake(dependency.__name__), dependency, True
            )

        logger.trace(f"Resolving {dependency} for container {self}")

        existing = self.get_provider(dependency)
        if existing:
            logger.trace(f"Found existing {existing}")
            return existing.get_instance(dependency, self)

        logger.trace(f"Existing not found, auto-wiring {dependency}")

        # region list injection special case
        # check if the dependency type is a list
        sequence_type, element_type = get_sequence_type(dependency.type)
        if element_type is not None and not _is_illegal_type(element_type):
            element_type: Any
            element_dependency = Dependency(dependency.name, element_type, True)
            elements = []
            for provider in self.get_providers(element_dependency):
                elements.append(provider.get_instance(element_dependency, self))

            if len(elements) > 0:
                return sequence_type(elements)

        # endregion

        result = self.autowire(dependency.type)

        self.add(
            Provider.from_supplier(lambda: result, dependency.type, dependency.name)
        )

        logger.trace(f"Successfully autowired {dependency} to {result}")
        return result

    def autowire(
            self,
            t: Type[_T],
            **explicit_kw_args,
    ) -> _T:
        """
        Auto-wires an object of the given type. Meaning that all dependencies of the object are resolved
        as defined by `self.resolve(...)` and the object is initialized with the resolved dependencies.
        In contrast to `self.resolve(...)`, this function does not store the result as a singleton provider.

        :param t:
        :param explicit_kw_args:
        :return: The auto-wired object
        :raises AutowiredException: if the object cannot be auto-wired
        """
        logger.trace(f"Auto-wiring {t} with {len(explicit_kw_args)} explicit args")
        if _is_illegal_type(t):
            raise IllegalAutoWireType(f"Cannot auto-wire object of type {t}")

        dependencies = _get_dependencies_for_type(t)

        resolved_kw_args = dict(explicit_kw_args) if explicit_kw_args else {}

        for dep in dependencies:
            if dep.name in resolved_kw_args:
                continue

            existing = self.get_provider(dep)
            if existing:
                logger.trace(f"Found existing {existing} provider for {dep}")

                resolved_kw_args[dep.name] = existing.get_instance(dep, self)
            elif dep.default_factory is not None:
                logger.trace(f"Using default factory for {dep}")
                resolved_kw_args[dep.name] = dep.default_factory()
            else:
                # try to resolve dependency
                try:
                    auto = self.resolve(dep)
                    resolved_kw_args[dep.name] = auto
                except AutowiredException as e:
                    if dep.required:
                        raise UnresolvableDependencyException(
                            f"Failed to resolve dependency {dep.name} "
                            f"of type {dep.type} for {t}. "
                        ) from e

        try:
            return t(**resolved_kw_args)
        except Exception as e:
            raise InstantiationError(f"Failed to initialize {t.__name__}") from e

    # use component scan to add providers by type

    def component_scan(self, root_module: Module) -> None:
        """
        Scans the given module and all submodules for classes annotated with `@component`.
        Adds a singleton provider for each class found.

        :param root_module: The root module to scan
        """
        for component_info in component_scan(root_module):
            provider = Provider.from_class(
                component_info.cls, self, component_info.transient
            )
            self.add(provider)


# region utils

def _instance_to_supplier(instance: Any) -> Callable[[], Any]:
    return lambda: instance


def _get_dependencies_for_type(t: type) -> List[Dependency]:
    init = _get_actual_init(t)
    dependencies = []
    if init:
        if dataclasses.is_dataclass(t):
            # noinspection PyDataclass
            for field in dataclasses.fields(t):
                annotation = field.type

                if field.default != dataclasses.MISSING:
                    default_factory = _instance_to_supplier(field.default)
                elif field.default_factory != dataclasses.MISSING:
                    default_factory = field.default_factory
                else:
                    default_factory = None

                dependency = Dependency(
                    field.name, annotation, field.default == dataclasses.MISSING, default_factory
                )
                dependencies.append(dependency)
        else:
            sig = inspect.signature(init)
            for name, param in sig.parameters.items():
                if name == "self":
                    continue
                annotation = param.annotation

                default = param.default
                has_default = default != inspect.Parameter.empty

                if annotation == inspect.Parameter.empty:
                    if has_default:
                        # falling back to the type of the default value
                        annotation = type(param.default)
                    else:
                        annotation = object

                dependency = Dependency(name, annotation, not has_default,
                                        _instance_to_supplier(default) if has_default else None)
                dependencies.append(dependency)

    return dependencies


def _camel_to_snake(name: str) -> str:
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def _group_by(key_fn: Callable[[Any], Any], items: List[Any]) -> Dict[Any, List[Any]]:
    result = {}
    for item in items:
        key = key_fn(item)
        if key not in result:
            result[key] = []
        result[key].append(item)
    return result


def _get_actual_init(t: type) -> Optional[FunctionType]:
    if "__init__" in t.__dict__:
        return getattr(t, "__init__")
    else:
        for base in t.__bases__:
            if base == object:
                continue
            init = _get_actual_init(base)
            if init:
                return init
        return None

# endregion



import dataclasses
from abc import ABC
from typing import List, Tuple

import pytest

import threading
from dataclasses import dataclass
from unittest.mock import Mock

from autowired import (
    autowired,
    cached_property,
    Context,
    Container,
    Dependency,
    UnresolvableDependencyException,
    AmbiguousDependencyException,
    IllegalAutoWireType,
    InstantiationError,
    Provider,
    NotProvidedException,
    IllegalContextClass,
    MissingTypeAnnotation,
    provided,
    thread_local_cached_property,
)


class Service0:
    pass


@dataclass
class Service1:
    service0: Service0


class Service2:
    def __init__(self, service1: Service1, blabla: str = "blabla"):
        self.service1 = service1
        self.blabla = blabla


@dataclass
class Service3:
    service0: Service0
    service1: Service1
    foo: str = "bar"


##


class ExampleContext(Context):
    service2: Service2 = autowired()

    @cached_property
    def service1(self) -> Service1:
        return self.autowire(Service1)

    @cached_property
    def service3(self) -> Service3:
        return self.autowire(Service3, foo="baz")


def test_service_instances():
    ctx = ExampleContext()

    service2 = ctx.service2
    service3 = ctx.service3
    service2b = ctx.service2

    # Assert that the same instance is returned by the context
    assert id(service2) == id(service2b), "Different instances of service2"

    # Assert that property `foo` is overridden
    assert service3.foo == "baz", "Incorrect foo value"

    # Assert that service2 in service3 and service2 are same
    assert id(service3.service1) == id(ctx.service1)

    assert isinstance(ctx.service1, Service1)
    assert isinstance(ctx.service2, Service2)
    assert isinstance(ctx.service3, Service3)


@dataclass
class ServiceX:
    s: str


@dataclass
class Service4:
    x: ServiceX
    service1: Service1


def test_derived_context():
    ctx = ExampleContext()

    class DerivedContext(Context):
        service4: Service4 = autowired()

        def __init__(self, parent_context: Context):
            self.derive_from(parent_context)

    derived_context = DerivedContext(ctx)
    derived_context.container.add(Provider.from_instance(ServiceX("foo")))
    assert isinstance(derived_context.service4, Service4)
    assert id(derived_context.service4.service1) == id(ctx.service1)
    assert derived_context.service4.x.s == "foo"


def test_unresolvable_dependency():
    class TestContext(Context):
        service4: Service4 = autowired()

    ctx = TestContext()

    with pytest.raises(UnresolvableDependencyException):
        print(ctx.service4)


def test_context_provider_names():
    class TestContext(Context):
        service1: Service1 = autowired()
        my_service_2: Service2 = autowired()

    ctx = TestContext()
    providers = ctx.container.get_providers()
    assert len(providers) == 2

    assert "service1" in [c.get_name() for c in providers]
    assert "my_service_2" in [c.get_name() for c in providers]

    class PascalCaseToSnakeCase:
        pass

    container = Container()
    container.add(PascalCaseToSnakeCase())
    providers = container.get_providers()
    assert len(providers) == 1
    assert providers[0].get_name() == "pascal_case_to_snake_case"


def test_ambiguous_dependency():
    class TestContext(Context):
        service1_a: Service1 = autowired()
        service1_b: Service1 = autowired()

    ctx = TestContext()

    with pytest.raises(AmbiguousDependencyException):
        ctx.autowire(Service2)

    with pytest.raises(AmbiguousDependencyException):
        ctx.container.resolve(Dependency("service1", Service1, True))

    s1a = ctx.container.resolve(Dependency("service1_a", Service1, True))
    s1b = ctx.container.resolve(Dependency("service1_b", Service1, True))

    assert isinstance(s1a, Service1)
    assert isinstance(s1b, Service1)
    assert id(s1a) != id(s1b)


def test_not_instantiable():
    class TestContext(Context):
        service_x: ServiceX = autowired()

    ctx = TestContext()
    try:
        ctx.service_x
    except Exception as e:
        assert isinstance(e, UnresolvableDependencyException)
        cause = e.__cause__
        assert isinstance(cause, IllegalAutoWireType)


def test_resolve_by_type():
    ctx = ExampleContext()
    service1 = ctx.container.resolve(Service1)

    assert id(ctx.service1) == id(service1)


def test_eager_init():
    names_initiated = []

    class TestService:
        def __init__(self, name: str):
            nonlocal names_initiated
            names_initiated.append(name)
            self.name = name

    class TestContext(Context):
        service1: TestService = autowired(eager=True, name="service1")
        service2: TestService = autowired(eager=False, name="service2")

    ctx = TestContext()

    assert len(names_initiated) == 1
    assert names_initiated[0] == "service1"

    service1 = ctx.service1

    assert len(names_initiated) == 1

    service2 = ctx.service2

    assert len(names_initiated) == 2

    assert service1.name == "service1"
    assert service2.name == "service2"


def test_autowire_initialization_error():
    container = Container()

    class TestService:
        def __init__(self):
            raise Exception("Test exception")

    with pytest.raises(InstantiationError):
        container.autowire(TestService)


def test_unregister_dependency():
    container = Container()
    s0 = Service0()
    container.add(Provider.from_instance(s0))

    s0_resolved = container.resolve(Service0)

    assert id(s0) == id(s0_resolved)

    container.remove("service0")

    s0_resolved2 = container.resolve(Service0)

    assert id(s0) != id(s0_resolved2)


def test_conflicting_bean():
    container = Container()

    class Component:
        pass

    container.add(Component())
    container.add(Component())

    named = Component()
    container.add(Provider.from_instance(named, "named_component"))

    with pytest.raises(AmbiguousDependencyException):
        container.resolve(Component)

    with pytest.raises(AmbiguousDependencyException):
        container.resolve(Dependency("component", Component))

    # should work with exact name
    container.resolve(Dependency("named_component", Component))

    container.add(Provider.from_instance(Component(), "named_component"))

    # should not work with multiple components with the same name
    with pytest.raises(AmbiguousDependencyException):
        container.resolve(Dependency("named_component", Component))


def test_use_correct_subtype():
    class ServiceA:
        pass

    class ServiceB(ServiceA):
        pass

    @dataclass
    class OtherService:
        a: ServiceA

    class TestContext(Context):
        other: OtherService = autowired()

    ctx = TestContext()
    service_b = ServiceB()
    ctx.container.add(Provider.from_instance(service_b))

    assert isinstance(ctx.other.a, ServiceB)
    assert id(ctx.other.a) == id(service_b)

    # same test with cached_property instead of register
    class TestContext2(Context):
        other: OtherService = autowired()

        @cached_property
        def service_b(self) -> ServiceB:
            return self.autowire(ServiceB)

    ctx2 = TestContext2()

    assert isinstance(ctx2.other.a, ServiceB)
    assert id(ctx2.other.a) == id(ctx2.service_b)


def test_provided():
    class TestContext(Context):
        service1: Service1 = provided()
        service2: Service2 = autowired()

        def __init__(self, service1: Service1 = None):
            if service1 is not None:
                self.service1 = service1

    with pytest.raises(NotProvidedException):
        TestContext()

    ctx = TestContext(Service1(Service0()))

    assert isinstance(ctx.service1, Service1)

    assert id(ctx.service1) == id(ctx.service2.service1)


def test_dataclass_as_context():
    @dataclass
    class TestContext(Context):
        service1: Service1 = autowired()

    with pytest.raises(IllegalContextClass):
        TestContext()


def test_missing_type_annotation():
    class TestContext(Context):
        service1 = autowired()

    ctx = TestContext()

    with pytest.raises(MissingTypeAnnotation):
        print(ctx.service1)


def test_register_instance():
    class TestContext(Context):
        service1: Service1 = autowired()
        service4: Service4 = autowired()

    ctx = TestContext()

    with pytest.raises(UnresolvableDependencyException):
        print(ctx.service4)

    x = ServiceX("foo")
    ctx.container.add(x)

    assert isinstance(ctx.service4, Service4)
    assert id(ctx.service4.x) == id(x)


def test_property_not_annotated():
    class TestContext1(Context):
        service1 = autowired()

    ctx = TestContext1()

    with pytest.raises(MissingTypeAnnotation):
        print(ctx.service1)

    class TestContext2(Context):
        @cached_property
        def service1(self):
            return self.autowire(Service1)

    with pytest.raises(MissingTypeAnnotation):
        print(TestContext2().container)

    class TestContext3(Context):
        # partially annotated
        service1: Service1 = autowired()
        service2 = autowired()

    with pytest.raises(MissingTypeAnnotation):
        print(TestContext3().container)


def test_provider_reuses_base_init():
    class Dep:
        pass

    class BaseProvider:
        def __init__(self, dep: Dep):
            self.dep = dep

    class Provider(BaseProvider):
        pass

    class TestContext(Context):
        provider: Provider = autowired()

    ctx = TestContext()
    assert isinstance(ctx.provider.dep, Dep)


def test_hidden_provider_is_singleton():
    class HiddenDep:
        pass

    class A:
        def __init__(self, hidden_dep: HiddenDep):
            self.hidden_dep = hidden_dep

    class B:
        def __init__(self, hidden_dep: HiddenDep):
            self.hidden_dep = hidden_dep

    class TestContext(Context):
        a: A = autowired()
        b: B = autowired()

    ctx = TestContext()

    assert id(ctx.a.hidden_dep) == id(ctx.b.hidden_dep)


def test_property_with_underscore():
    class TestContext(Context):
        _service1: Service1 = autowired()

        @cached_property
        def _service2(self) -> Service2:
            return self.autowire(Service2)

    ctx = TestContext()

    providers = ctx.container.get_providers()
    assert len(providers) == 2

    provider_names = [c.get_name() for c in providers]

    assert "service1" in provider_names
    assert "service2" in provider_names


def test_singletons():
    class ServiceA:
        pass

    class ServiceB:
        pass

    class ServiceC:
        def __init__(self, a: ServiceA, b: ServiceB):
            self.a = a
            self.b = b

    class ServiceD:
        def __init__(self, a: ServiceA, b: ServiceB):
            self.a = a
            self.b = b

    class ServiceE:
        pass

    class ServiceF:
        pass

    class TestContext(Context):
        service_c: ServiceC = autowired()
        service_d: ServiceD = autowired()
        service_f: ServiceF = provided()

        def __init__(self, service_f: ServiceF):
            self.service_f = service_f

        @property
        def service_b(self) -> ServiceB:
            return self.autowire(ServiceB)

        @cached_property
        def service_e(self) -> ServiceE:
            return self.autowire(ServiceE)

    f = ServiceF()
    ctx = TestContext(f)

    # property should not be singleton
    assert id(ctx.service_c.b) != id(ctx.service_d.b)
    # autowired should be singleton
    assert id(ctx.service_c) == id(ctx.service_c)
    # cached_property should be singleton
    assert id(ctx.service_e) == id(ctx.service_e)
    # provided should be singleton
    assert id(ctx.service_f) == id(ctx.service_f)


def test_property_fields():
    class ServiceA:
        pass

    class ServiceB:
        pass

    @dataclass
    class ServiceC:
        a: ServiceA
        b: ServiceB

    class ServiceD:
        def __init__(self, a: ServiceA, b: ServiceB):
            self.a = a
            self.b = b

    class TestContext(Context):
        service_a: ServiceA = cached_property(lambda self: self.autowire(ServiceA))
        service_b: ServiceB = property(lambda self: self.autowire(ServiceB))

        service_c: ServiceC = autowired()
        service_d: ServiceD = autowired()

    ctx = TestContext()

    assert isinstance(ctx.service_a, ServiceA)
    assert isinstance(ctx.service_b, ServiceB)
    assert isinstance(ctx.service_c, ServiceC)
    assert isinstance(ctx.service_d, ServiceD)

    # cached_property should be singleton
    assert id(ctx.service_c.a) == id(ctx.service_a)
    # property should not be singleton
    assert id(ctx.service_c.b) != id(ctx.service_b)


def test_transient_autowired_field():
    class ServiceA:
        pass

    class ServiceB:
        pass

    class ServiceC:
        def __init__(self, a: ServiceA, b: ServiceB):
            self.a = a
            self.b = b

    class ServiceD:
        def __init__(self, a: ServiceA, b: ServiceB):
            self.a = a
            self.b = b

    class TestContext(Context):
        service_a: ServiceA = autowired()
        service_b: ServiceB = autowired(transient=True)
        service_c: ServiceC = autowired()
        service_d: ServiceD = autowired()

    ctx = TestContext()

    assert id(ctx.service_c.b) != id(ctx.service_d.b)
    assert id(ctx.service_c.a) == id(ctx.service_d.a)

    assert id(ctx.service_b) != id(ctx.service_b)
    assert id(ctx.service_a) == id(ctx.service_a)


def test_untyped_component_arg():
    class ServiceA:
        def __init__(self, arg):
            self.arg = arg

    container = Container()

    with pytest.raises(UnresolvableDependencyException):
        container.autowire(ServiceA)


def test_untyped_component_arg_with_default():
    class SomeClass:
        pass

    default_value = SomeClass()

    class SomeService:
        def __init__(self, arg=default_value):
            self.arg = arg

    container = Container()

    not_default_value = SomeClass()

    container.add(not_default_value)
    service = container.resolve(SomeService)

    assert isinstance(service.arg, SomeClass)
    assert service.arg is not default_value
    assert service.arg is not_default_value


def test_inherited_context():
    class ServiceA:
        def __init__(self, arg: str = ""):
            self.arg = arg

    class ServiceB:
        def __init__(self, arg: str = ""):
            self.arg = arg

    class BaseContext(Context):
        service_a: ServiceA = autowired()

    class ExtendedContext(BaseContext):
        service_b: ServiceB = autowired()

    ctx = ExtendedContext()

    assert isinstance(ctx.service_a, ServiceA)
    assert isinstance(ctx.service_b, ServiceB)

    class ExtendedContext2(BaseContext):
        service_a: ServiceA = autowired(arg="foo")

        @cached_property
        def service_b(self) -> ServiceB:
            return self.autowire(ServiceB, arg="bar")

    ctx2 = ExtendedContext2()

    assert ctx2.service_a.arg == "foo"
    assert ctx2.service_b.arg == "bar"


def test_provider_from_supplier():
    instance = Service0()

    def singleton_supplier():
        return instance

    container = Container()
    container.add(Provider.from_supplier(singleton_supplier, Service0))

    assert container.resolve(Service0) is instance
    assert container.resolve(Service0) is container.resolve(Service0)

    # provider should be removed by equality
    def singleton_supplier2():
        return instance

    # should not remove anything
    container.remove(Provider.from_supplier(singleton_supplier2, Service0))
    assert len(container.get_providers()) == 1
    # should remove the provider
    container.remove(Provider.from_supplier(singleton_supplier, Service0))
    assert len(container.get_providers()) == 0

    container = Container()

    def transient_supplier():
        return Service0()

    container.add(Provider.from_supplier(transient_supplier, Service0))

    assert container.resolve(Service0) is not container.resolve(Service0)


def test_provider_factory_methods():
    singleton_provider = Provider.from_instance(Service0())

    assert isinstance(singleton_provider.get_instance(Mock(), Mock()), Service0)
    assert singleton_provider.get_instance(
        Mock(), Mock()
    ) is singleton_provider.get_instance(Mock(), Mock())

    provider_from_type = Provider.from_supplier(Service0)
    assert isinstance(provider_from_type.get_instance(Mock(), Mock()), Service0)
    assert provider_from_type.get_instance(
        Mock(), Mock()
    ) is not provider_from_type.get_instance(Mock(), Mock())

    def supplier_untyped():
        return Service0()

    with pytest.raises(MissingTypeAnnotation):
        Provider.from_supplier(supplier_untyped)

    assert isinstance(
        Provider.from_supplier(supplier_untyped, Service0).get_instance(Mock(), Mock()),
        Service0,
    )

    def supplier_typed() -> Service0:
        return Service0()

    assert isinstance(
        Provider.from_supplier(supplier_typed).get_instance(Mock(), Mock()),
        Service0,
    )


def test_thread_local_property():
    import threading

    class ServiceA:
        def __init__(self):
            self.thread_id = threading.get_ident()

    class ServiceB:
        def __init__(self, service_a: ServiceA):
            self.service_a = service_a
            self.thread_id = threading.get_ident()

    class TestContext(Context):
        service_a: ServiceA = autowired()

        @thread_local_cached_property
        def service_b(self) -> ServiceB:
            return self.autowire(ServiceB)

    ctx = TestContext()
    main_thread_service_b = ctx.service_b

    assert isinstance(main_thread_service_b, ServiceB)
    assert main_thread_service_b is ctx.service_b

    def in_new_thread():
        new_thread_service_b = ctx.service_b
        assert new_thread_service_b is not main_thread_service_b
        assert new_thread_service_b is ctx.service_b

        assert new_thread_service_b.service_a is main_thread_service_b.service_a

    _test_in_thread(in_new_thread)


def _test_in_thread(func):
    thread_error = None

    def in_new_thread():
        try:
            func()
        except Exception as e:
            nonlocal thread_error
            thread_error = e

    thread = threading.Thread(target=in_new_thread)
    thread.start()
    thread.join()

    if thread_error is not None:
        raise thread_error


def test_thread_local_autowired():
    class TestContext(Context):
        service1: Service1 = autowired(thread_local=True)

    ctx = TestContext()
    main_thread_service1 = ctx.service1
    assert isinstance(main_thread_service1, Service1)
    assert ctx.service1 is main_thread_service1

    def in_new_thread():
        assert isinstance(ctx.service1, Service1)
        assert ctx.service1 is not main_thread_service1
        assert ctx.service1 is ctx.service1

    _test_in_thread(in_new_thread)


def test_thread_safe_autowired_field():
    class SlowInitService:
        def __init__(self):
            import time

            time.sleep(0.1)

    class MyContext(Context):
        service: SlowInitService = autowired()

    ctx = MyContext()

    from_thread = None

    def in_thread():
        nonlocal from_thread
        from_thread = ctx.service

    thread = threading.Thread(target=in_thread)
    thread.start()

    from_main = ctx.service

    thread.join()
    assert from_main is from_thread


def test_context_value_selector():
    @dataclass
    class ServiceA:
        foo: str = "bar"

    @dataclass
    class ServiceB:
        foo: str = "baz"

    @dataclass
    class ServiceC:
        foo: object

    @dataclass
    class ServiceD:
        foo: object

    @dataclass
    class InnerConfig:
        foo: str

    class Config:
        def __init__(self, inner: InnerConfig):
            self.inner = inner

    class MyContext(Context):
        config: Config = provided()
        a: ServiceA = autowired(foo=config.inner.foo)
        # noinspection PyUnresolvedReferences
        b: ServiceB = autowired(foo=config.wrong)
        c: ServiceC = autowired(foo=object())
        d: ServiceD = autowired(foo=c.foo)

        def __init__(self, config: Config):
            self.config = config

    ctx = MyContext(Config(InnerConfig("bar")))

    assert ctx.a.foo == "bar"

    with pytest.raises(AttributeError):
        print(ctx.b)

    assert ctx.d.foo is ctx.c.foo


def test_context_value_selector_direct():
    @dataclass
    class TestService:
        service1: Service1
        foo: str

    class TestContext(Context):
        foo: str = provided()
        service1: Service1 = autowired()
        test_service: TestService = autowired(service1=service1, foo=foo)

        def __init__(self):
            self.foo = "bar"

    ctx = TestContext()
    assert isinstance(ctx.test_service, TestService)
    assert ctx.test_service.service1 is ctx.service1
    assert ctx.test_service.foo == "bar"


def test_generic_dependency():
    @dataclass
    class SomeData:
        pass

    @dataclass
    class Service:
        data: List[SomeData]

    class TestContext(Context):
        service: Service = autowired()

        @cached_property
        def data(self) -> List[SomeData]:
            return [SomeData()]

    ctx = TestContext()
    assert isinstance(ctx.service, Service)
    for data in ctx.service.data:
        assert isinstance(data, SomeData)


def test_list_injection():
    class Plugin(ABC):
        pass

    @dataclass
    class PluginService:
        plugins: List[Plugin]

    class PluginA(Plugin):
        pass

    class PluginB(Plugin):
        pass

    def plugin_container():
        container = Container()
        container.add(PluginA())
        container.add(PluginB())
        return container

    def assert_plugin_service(plugin_service, sequence_type):
        assert isinstance(plugin_service, PluginService)
        assert len(plugin_service.plugins) == 2
        assert isinstance(plugin_service.plugins, sequence_type)
        assert isinstance(plugin_service.plugins[0], PluginA)
        assert isinstance(plugin_service.plugins[1], PluginB)

    container = plugin_container()

    plugin_service = container.resolve(PluginService)

    assert_plugin_service(plugin_service, list)

    # test with tuple

    @dataclass
    class PluginService:
        plugins: Tuple[Plugin, ...]

    container = plugin_container()

    plugin_service = container.resolve(PluginService)
    assert_plugin_service(plugin_service, tuple)

    # test illegal tuple type
    @dataclass
    class PluginService:
        plugins: Tuple[Plugin]

    container = plugin_container()

    with pytest.raises(UnresolvableDependencyException):
        container.resolve(PluginService)


def test_list_injection_no_existing():
    @dataclass(frozen=True)
    class Value:
        v: int

    @dataclass
    class SomeData:
        some_list: List[Value] = dataclasses.field(default_factory=lambda: [Value(0), Value(1)])

    container = Container()

    some_data = container.resolve(SomeData)

    assert isinstance(some_data, SomeData)
    assert isinstance(some_data.some_list, list)
    # should use default if no value components are available
    assert some_data.some_list == [Value(0), Value(1)]

    container2 = Container()
    container2.add(Provider.from_instance(Value(1)))
    container2.add(Provider.from_instance(Value(2)))

    some_data2 = container2.resolve(SomeData)

    assert isinstance(some_data2, SomeData)
    assert isinstance(some_data2.some_list, list)
    # default should have priority even if value components are available
    assert some_data2.some_list == [Value(0), Value(1)]

    @dataclass
    class SomeOtherData:
        some_list_without_default: List[Value]

    container3 = Container()
    container3.add(Provider.from_instance(Value(1)))
    container3.add(Provider.from_instance(Value(2)))
    container3.add(Provider.from_instance(Value(3)))
    container3.add(Provider.from_instance(Value(4)))

    some_other_data = container3.resolve(SomeOtherData)

    assert isinstance(some_other_data, SomeOtherData)
    assert isinstance(some_other_data.some_list_without_default, list)
    # no default, so should use all available value components
    assert some_other_data.some_list_without_default == [
        Value(1),
        Value(2),
        Value(3),
        Value(4),
    ]

    container4 = Container()
    # no defaults and no value components: should fail
    with pytest.raises(UnresolvableDependencyException):
        container4.resolve(SomeOtherData)

    @dataclass
    class SomeDataWithIntList:
        some_list: List[int]

    container5 = Container()
    container5.add(Provider.from_instance(1))
    container5.add(Provider.from_instance(2))

    # list injection not for primitive types
    with pytest.raises(UnresolvableDependencyException):
        container5.resolve(SomeDataWithIntList)



import sys

import pytest
from typing import List, Dict, Tuple, Union, Any, Set

# noinspection PyProtectedMember
from autowired._typing_utils import is_subtype, get_list_element_type


def test_non_generic_types():
    assert is_subtype(int, object) is True
    assert is_subtype(bool, int) is True
    assert is_subtype(Union[int, str], object) is True
    assert is_subtype(int, Union[int, str]) is True
    assert is_subtype(int, str) is False


def test_generic_with_non_generic():
    assert is_subtype(List[int], List) is True
    assert is_subtype(Dict, Dict[Any, Any]) is True


def test_generic_types():
    assert is_subtype(List[int], List[int]) is True
    assert (
        is_subtype(List[int], List[object]) is True
    )  # Not type-safe, but expected by the function's behavior
    assert is_subtype(Dict[str, int], Dict[object, object]) is True  # Same as above
    assert is_subtype(Tuple[int, ...], Tuple[object, ...]) is True  # Same as above
    assert is_subtype(Dict[str, int], Dict[str, str]) is False


def test_generic_types_with_different_lengths():
    assert is_subtype(Tuple[int, int], Tuple) is True
    assert is_subtype(Tuple, Tuple[Any, ...]) is True


def test_variant_generic_types():
    # For fully correct subtype checks, these should be False, but the
    # implementation as it stands will return True.
    assert is_subtype(List[int], List[object]) is True
    assert is_subtype(Dict[str, int], Dict[Any, Any]) is True


def test_complex_generic_types():
    assert (
        is_subtype(List[Dict[str, int]], List[Dict[object, object]]) is True
    )  # Not type-safe due to variance, but expected by the function's behavior
    assert (
        is_subtype(
            Dict[Tuple[int, str], List[int]],
            Dict[Tuple[object, object], List[object]],
        )
        is True
    )  # Same as above


def test_any_type():
    assert is_subtype(Any, int) is True
    assert is_subtype(int, Any) is True
    assert is_subtype(Any, List[int]) is True
    assert is_subtype(List[int], Any) is True
    assert is_subtype(Any, List) is True
    assert is_subtype(List, Any) is True
    assert is_subtype(Any, Any) is True
    assert is_subtype(List[int], List[Any]) is True
    assert is_subtype(Dict[Any, Any], Dict) is True


def test_inheritance():
    class MyBase:
        pass

    class MyDerived(MyBase):
        pass

    assert is_subtype(MyDerived, MyBase) is True
    assert is_subtype(MyBase, MyDerived) is False
    assert (
        is_subtype(List[MyDerived], List[MyBase]) is True
    )  # Not type-safe, but expected by the function's behavior


def test_mixed_generic_non_generic():
    assert is_subtype(List[int], List) is True
    assert is_subtype(List, List[int]) is True
    assert is_subtype(List[int], list) is True
    assert is_subtype(list, List[int]) is True

    assert is_subtype(List[int], Set) is False
    assert is_subtype(Set, List[int]) is False
    assert is_subtype(List[int], set) is False
    assert is_subtype(set, List[int]) is False


_version_info = sys.version_info
if _version_info.major >= 3 and _version_info.minor >= 10:
    new_union_syntax = [True, False]
else:
    new_union_syntax = [False]


@pytest.mark.parametrize("new_union_syntax", new_union_syntax)
def test_union_types(new_union_syntax: bool):
    def union(*args):
        if new_union_syntax:
            if len(args) == 1:
                return args[0]
            if len(args) == 2:
                return args[0] | args[1]
            if len(args) == 3:
                return args[0] | args[1] | args[2]
        else:
            return Union[args]

    # case 1: both types are unions
    # 1.1: same types
    assert is_subtype(union(int, str), union(int, str)) is True
    assert is_subtype(union(int, str), union(str, int)) is True
    # 1.2: all types in left are subtypes of at least one type in right
    assert is_subtype(union(int, str), union(object, str)) is True
    assert is_subtype(union(int, str), union(str, object)) is True
    # 1.3: not all types in left are subtypes of at least one type in right
    assert is_subtype(union(int, str), union(float, str)) is False
    assert is_subtype(union(int, str), union(str, float)) is False

    assert is_subtype(union(int, str), Any) is True

    # case 2: only the left type is union
    assert is_subtype(union(int, str), int) is False
    assert is_subtype(union(int, str), str) is False
    assert is_subtype(union(int, str), object) is True

    # case 3: only the right type is union
    assert is_subtype(int, union(int, str)) is True
    assert is_subtype(str, union(int, str)) is True
    assert is_subtype(float, union(int, str)) is False

    # case 4: one is Any
    assert is_subtype(Any, union(int, str)) is True
    assert is_subtype(union(int, str), Any) is True


def test_get_list_type():
    assert get_list_element_type(List[int]) == int
    assert get_list_element_type(List) is None
    assert get_list_element_type(List[object]) is object
    assert get_list_element_type(int) is None



from typing import Type, get_args, get_origin, Union, Any, Optional, List, Tuple

try:
    from types import UnionType
except ImportError:  # pragma: no cover
    UnionType = None


def is_subtype(t1: Type, t2: Type) -> bool:
    """
    Checks if t1 is a subtype of t2 (instances of t1 can be used where instances of t2 are expected).
    Similar to issubclass, but also works for generic types.

    Note that this is a simple implementation that does not take invariant type arguments into account.
    Meaning is_subtype(List[int], List[object]) will return True, although strictly speaking
    List[int] is not a subtype of List[object], since it is a mutable container and therefore invariant.

    :param t1:
    :param t2:
    :return:
    """

    if t1 is t2:
        return True

    # region union type support
    # union type similarity check rule: all types of t1 must be subtypes of at least one type of t2
    t1_union_types = _as_union_types(t1)
    t2_union_types = _as_union_types(t2)

    if len(t1_union_types) > 1 or len(t2_union_types) > 1:
        return all(
            any(is_subtype(t1_arg, t2_arg) for t2_arg in t2_union_types)
            for t1_arg in t1_union_types
        )
    # endregion

    if t1 is Any or t2 is Any:
        return True

    # both types are not generic -> we can use issubclass
    if get_origin(t1) is None and get_origin(t2) is None:
        return issubclass(t1, t2)

    origin1 = get_origin(t1) or t1
    origin2 = get_origin(t2) or t2

    # base condition: t1 must be a subclass of t2, otherwise we can already return False
    if not issubclass(origin1, origin2):
        return False

    # from now on t1 is a subclass of t2
    # -> we only need to check type arguments now

    # only the one type is generic -> we consider the argument to be Any
    #   -> t1 = t1[Any, Any, ...] and t2 = t2[x, y, ...]
    #   or t1 = t1[x, y, ...] and t2 = t2[Any, Any, ...]
    if get_origin(t1) is None or get_origin(t2) is None:
        return True

    args1 = get_args(t1)
    args2 = get_args(t2)

    # if one of the types has no type arguments, same as above
    if not args1 or not args2:
        return True

    # compare each of the type arguments recursively
    # as above,
    for arg1, arg2 in zip(args1, args2):
        if arg1 is Ellipsis or arg2 is Ellipsis:
            # again, handle as Any
            continue
        if not is_subtype(arg1, arg2):
            return False

    return True


def _as_union_types(t) -> Tuple[Type, ...]:
    """
    Returns the types of a Union type, or a tuple containing only t if t is not a Union type.
    """

    if get_origin(t) is Union:
        return get_args(t)

    if UnionType is not None:
        if get_origin(t) is UnionType:
            return get_args(t)

    return (t,)


def get_list_element_type(t: Type) -> Optional[Type]:
    """
    Returns the type of the elements of a list type, or None if t is not a list type.
    """
    origin = get_origin(t)
    if origin is list or origin is List:
        args = get_args(t)
        if args:
            return args[0]
    return None


def get_sequence_type(t: Type) -> Union[Tuple[Type, Type], Tuple[None, None]]:
    """
    Returns the type of the elements of a list type, or None if t is not a list type.
    """
    origin = get_origin(t)
    if origin is list or origin is List:
        args = get_args(t)
        if args:
            return list, args[0]

    if origin is tuple or origin is Tuple:
        args = get_args(t)
        if len(args) == 2 and args[1] is Ellipsis:
            return tuple, args[0]

    return None, None



import dataclasses
import inspect
import threading
from abc import ABC
from collections import defaultdict
from functools import cached_property
from typing import Optional, Union, Any, Callable, Dict, Type, List, TypeVar
from dataclasses import dataclass
from ._container import Provider, Container
from ._exceptions import (
    MissingTypeAnnotation,
    IllegalContextClass,
    NotProvidedException,
)
from ._logging import logger
from ._thread_local_cached_property import thread_local_cached_property


_T = TypeVar("_T")


class _ContextValueSelector:
    def __init__(
        self,
        parent: Optional["_ContextValueSelector"] = None,
        key: Optional[str] = None,
    ):
        self.parent = parent
        self.key = key

    def __getattr__(self, item: str) -> "_ContextValueSelector":
        return _ContextValueSelector(self, item)

    def select(self, ctx: "Context") -> Any:
        assert self.parent is not None
        assert self.key is not None

        parent_value = self.parent.select(ctx)
        return getattr(parent_value, self.key)


class _ContextProperty(ABC, _ContextValueSelector):
    name: str = None

    def __setattr__(self, key, value):
        if key == "name":
            assert self.name is None, f"Cannot set {self.__class__.__name__} name twice"
        super().__setattr__(key, value)

    def select(self, ctx: "Context") -> Any:
        assert self.name is not None
        return getattr(ctx, self.name)


class _Autowired(_ContextProperty):
    def __init__(
        self,
        eager: bool,
        transient: bool,
        thread_local: bool,
        kw_args_factory: Optional[Callable[["Context"], Dict[str, Any]]],
        kw_args: Dict[str, Any],
    ):
        super().__init__()
        self.eager = eager
        self.transient = transient
        self.thread_local = thread_local
        self.kw_args_factory = kw_args_factory
        self.kw_args = kw_args

    def get_all_kw_args(self, ctx: "Context") -> Dict[str, Any]:
        explicit_kw_args = self.kw_args_factory(ctx) if self.kw_args_factory else {}
        explicit_kw_args.update(self.kw_args)

        for key, value in explicit_kw_args.items():
            if isinstance(value, _ContextValueSelector):
                explicit_kw_args[key] = value.select(ctx)
        return explicit_kw_args

    def __getattr__(self, item):
        return _ContextValueSelector(self, item)


class _Provided(_ContextProperty):
    def __getattr__(self, item):
        return _ContextValueSelector(self, item)


def autowired(
    kw_args_factory: Optional[Callable[["Context"], Dict[str, Any]]] = None,
    /,
    *,
    eager: bool = False,
    transient: bool = False,
    thread_local: bool = False,
    **kw_args,
) -> Any:
    """
    Marks a context field as autowired.
    Auto-wired fields are converted to cached properties on the class.
    :param kw_args_factory: Return a dict of keyword arguments for initialization of the field
    :param eager: Eagerly initialize the field on object creation
    :param transient: Every access to the field returns a new instance
    :param thread_local: Create on instance per-thread
    :param kw_args: keyword arguments for initialization of the field
    :return:
    """
    return _Autowired(
        eager=eager,
        transient=transient,
        thread_local=thread_local,
        kw_args_factory=kw_args_factory,
        kw_args=kw_args,
    )


def provided() -> Any:
    """
    Marks a field as provided.
    Meaning that the field is set explicitly rather than auto-wired.
    If not set, an exception is raised on context initialization.
    :return:
    """
    return _Provided()


class _ContextMeta(type):
    """
    Metaclass for Context classes.
    Responsible for evaluating context properties (autowired and provided fields).
    """

    def __new__(mcs, name, bases, class_dict):
        autowired_fields = [
            key for key, value in class_dict.items() if isinstance(value, _Autowired)
        ]
        eager_fields = [
            key
            for key, value in class_dict.items()
            if isinstance(value, _Autowired) and value.eager
        ]
        provided_fields = [
            key for key, value in class_dict.items() if isinstance(value, _Provided)
        ]
        for key, value in class_dict.items():
            if isinstance(value, _ContextProperty):
                value.name = key

        original_init = class_dict.get("__init__", None)

        def new_init(self, *args, **kwargs):
            self._autowire_locks = defaultdict(threading.Lock)
            for field in autowired_fields:
                self._autowire_locks[field] = threading.Lock()

            if dataclasses.is_dataclass(self):
                raise IllegalContextClass(
                    f"Context class {name} must not be a dataclass"
                )
            if original_init:
                original_init(self, *args, **kwargs)

            for field_name in eager_fields:
                getattr(self, field_name)

            for field_name in provided_fields:
                value = getattr(self, field_name)
                if isinstance(value, _Provided):
                    raise NotProvidedException(
                        f"Field {field_name} is marked as provided but is not initialized"
                    )

        class_dict["__init__"] = new_init

        result = super().__new__(mcs, name, bases, class_dict)

        def __getattribute__(self, item):
            attr_value = super(result, self).__getattribute__(item)

            if not isinstance(attr_value, _Autowired):
                return attr_value

            field_type = _get_field_type(item, self)

            if attr_value.transient:
                value = self.autowire(field_type, **attr_value.get_all_kw_args(self))
            elif attr_value.thread_local:
                thread_local_storage = self.__dict__.setdefault(
                    f"_{item}_thread_local", threading.local()
                )

                if not hasattr(thread_local_storage, "value"):
                    thread_local_storage.value = self.autowire(
                        field_type, **attr_value.get_all_kw_args(self)
                    )

                value = thread_local_storage.value
            else:
                lock = self._autowire_locks[item]
                with lock:
                    attr_value = super(result, self).__getattribute__(item)
                    if isinstance(attr_value, _Autowired):
                        # is still an autowired field, auto-wire it
                        value = self.autowire(
                            field_type, **attr_value.get_all_kw_args(self)
                        )
                        self.__dict__[item] = value
                    else:
                        value = self.__dict__[item]

            return value

        result.__getattribute__ = __getattribute__

        return result


class Context(metaclass=_ContextMeta):
    def derive_from(self, ctx: Union["Context", Container]) -> None:
        """
        Registers all providers from the given context or container in this context.
        """
        container = ctx.container if isinstance(ctx, Context) else ctx
        for provider in container.get_providers():
            self.container.add(provider)

    @cached_property
    def container(self) -> Container:
        """
        Returns the container for this context.
        """
        container = Container()

        for prop in _get_obj_properties(self):
            if prop.name == "container":
                continue

            getter = _PropertyGetter(self, prop)
            if prop.type is None:
                raise MissingTypeAnnotation(
                    f"Failed to determine type of {type(self).__name__}.{prop.name}. "
                )

            name_normalized = prop.name.lstrip("_")

            container.add(Provider.from_supplier(getter, prop.type, name_normalized))

        return container

    def autowire(self, t: Union[Type[_T], type], **explicit_kw_args) -> _T:
        return self.container.autowire(t, **explicit_kw_args)


# region utils


@dataclass(frozen=True)
class _PropertyInfo:
    name: str
    type: Type[_T]


@dataclass
class _PropertyGetter(Callable[[], Any]):
    obj: Any
    property: _PropertyInfo

    def __call__(self) -> Any:
        logger.trace(f"Getting property {self.property.name} for {self.obj}")
        return getattr(self.obj, self.property.name)


def _get_annotations_for_type(t: type) -> Dict[str, Any]:
    annotations = {}

    if hasattr(t, "__annotations__"):
        annotations = t.__annotations__

    # base classes (do not overwrite from derived classes)
    for base in t.__bases__:
        if base == object:
            continue
        base_annotations = _get_annotations_for_type(base)
        annotations = {
            **base_annotations,
            **annotations,
        }

    return annotations


def _get_field_type(field_name: str, obj: Any) -> Type[_T]:
    annotations = _get_annotations_for_type(type(obj))
    field_type = annotations.get(field_name, None)
    if field_type is None:
        raise MissingTypeAnnotation(
            f"Cannot determine type of field {type(obj).__name__}.{field_name}"
        )

    return field_type


def _get_obj_properties(self) -> List[_PropertyInfo]:
    properties = []
    for name, attr in inspect.getmembers(type(self)):
        is_cached_property = isinstance(attr, cached_property)
        is_thread_local_cached_property = isinstance(attr, thread_local_cached_property)
        is_normal_property = isinstance(attr, property)

        if is_cached_property or is_normal_property or is_thread_local_cached_property:
            getter = attr.fget if is_normal_property else attr.func
            prop_type = getter.__annotations__.get("return", None)
            if prop_type is None and hasattr(self, "__annotations__"):
                # if no return type annotation is present,
                # try to get from class annotations
                prop_type = type(self).__annotations__.get(name, None)

            properties.append(_PropertyInfo(name, prop_type))
        elif isinstance(attr, _Autowired) or isinstance(attr, _Provided):
            properties.append(_PropertyInfo(name, _get_field_type(name, self)))

    return properties


# endregion



from autowired import component
from tests.component_scan_test_module import Controller
from tests.component_scan_test_module.test_package.file_module import (
    TestPackageFileComponentInitExposed,
)


@component
class TestPackageRootComponent(Controller):
    pass


class TestPackageRootNotComponent(Controller):
    pass
